/**
 * @file
 * @brief Global Definitions for the Netsweeper Squid URL Redirector
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#ifndef _REDIR_H
#define _REDIR_H

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <regex.h>

#define BUFFSIZE 8192

// Server thread state
#define	ST_DISCONNECTED	0
#define	ST_PEERDOWN	1

#define	ST_IDLE		2
#define ST_BUSY		3
#define	ST_WRTIMEOUT	4
#define	ST_RDTIMEOUT	5
#define ST_FATAL	6

typedef enum {
	SQUID_24 = 0,
	SQUID_25,
	SQUID_26,
	SQUID_27,
	SQUID_30,
	SQUID_31,
	SQUID_32
} protocol_select;



/** @brief The Request structure placed into the queue and consumed by a connection thread */
typedef struct _request {
	/** @brief id in squid request when concurrentcy is enabled */
	char*	id;

	/** @brief URL in Squid request */
	char*	URL;

	/** @brief actual request message sent to NSP */
	char*	message;

	/** @brief size of request message sent to NSP */
	int	size;

	/** times to live, retry times left for the request to be processed by NSP */
	int	ttl;

	/** if we are http based, used for fail close */
	int ishttp;

	/** @brief The next request when we are in the queue */
	struct _request* next;
} Request;

/** @brief Our global configuration */
typedef struct _global_conf {
	/** @brief defines the input protocol redir use to read from input stream. It may be SQUID_PROTOCOL_25, SQUID_PROTOCOL_26 ...*/
	protocol_select	input_protocol;

	/** @brief nsp port number for different protocols. It may be 3435, 3429, or 3428. 3435 is for squid25 by default.*/
	int	nsp_port;

	/** @brief the output protocol of redir to send to NSP. It may be SQUID_PROTOCOL_25, SQUID_PROTOCOL_26,... */
	protocol_select	output_protocol;

	/** @brief Flag used to indicate if we mangle 'domain\user' to 'user@domain' */
	int	username_domain_mangling;

	/** @brief Regular expression that defines which invalid characters will be replaced */
	regex_t* bad_username_chars_preg;

	/** @brief Flag used to indicate that we should turn username to lowercase */
	int	username_lowercase;

	/** @brief The domain name that should be removed from the username. If empty string remove any domain. */
	char*	username_remove_domain;
	size_t	username_remove_domain_len;

	/** @brief number of connections per server */
	int	ncons;

	/** @brief primary/secondary or round robin load balance type */
	int	load_bc_type;

	/** @brief max number of times to retry a request before it fails */
	int	max_retry_times;

	/** @brief connect timeout for redirector to communicate with NSP */
	int	connect_timeout;
	/** @brief read timeout for redirector to communicate with NSP */
	int	read_timeout;
	/** @brief write timeout for redirector to communicate with NSP */
	int	write_timeout;

	/** @brief read timeout for redir to communicate with Squid */
	int	squid_read_timeout;
	/** @brief write timeout for redir to communicate with Squid */
	int	squid_write_timeout;

	/** @brief between attempt to connect when the peer is down */
	int	reconnect_delay;

	/** @brief max timeout for each request */
	int	max_timeout;

	/** @brief max number of consequent timeout errors to shutdown connection */
	int	max_timeouts_num;

	/** @brief max number of charaters a squid request can hold */
	int	max_request_size;

	/** @brief max number of charaters a NSP replt can hold */
	int	max_reply_size;

	/** @brief max number of requests can be hold in queue */
	int	max_requests;

	/** @brief option to open/close request when filtering fails */
	int	fail_close;

	/** @brief url to return when filtering fails and the option fail_close is true */
	char*	fail_close_url;

	/** @brief the site:port we fail close to when the remote host is https or non-http */
	char* fail_close_host;

	/** @brief input file descriptor, default 0 for squid redirector helper */
	int	input_fd;

	/** @brief output file descriptor, default 1 for squid redirector helper */
	int	output_fd;

	/** @brief NSP host name:port */
	char**  nsp_hosts;

	/** @brief number of nsp hosts */
	int	nsp_num;

	/** @brief option to log to stderr */
	int	stderr_log;

	/** @brief log level */
	int	log_level;

	/** @brief flag that shows that policy server should support concurrent requests */
	int	server_concurrency;

	int	queue_wait_timeout;

	int	queue_flush_delay;

	struct simple_iphead *ip2domain;

	int     unescape;
} GlobalConf;

/** @brief The connection structure, used by a thread */
typedef struct _connection {
	/** @brief The connection thread ID */
	pthread_t thread_id;
	/** @brief The state of the connection, see ST_ */
	int	state;
	/** @brief The socket used for this connection */
	int	socket;

	/** @brief The hostname used for this connection */
	char	*hostname;

	/** @brief The IP Address for this host */
	in_addr_t  hostip;

	/** @brief The port used for this connection */
	int	port;

	/** @brief A pointer to the global configuration */
	GlobalConf*	global_conf;

	/** @brief Startup lock used when starting the thread/shutting down the thread */
	pthread_rwlock_t *start_lock;
} Connection;

void* server_thread(void *args);
int st_check_connection(Connection *conn);

void * xmalloc(size_t size);
char * xstrdup(char *str);


#endif
