/**
 * @file
 * @brief This file tests TCP connections openned by redir to nsp server
 *
 * One redir open maximum number of connections to One server
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

/* In these tests we error out if there isn't enough memory to run them.
 * We have to use system-specific mechanisms to find out how much memory
 * there is. On Linux we use sysinfo, and on Mac OS X we use sysctl.
 *
 * Unfortunately that means we have a rather ugly ifdef wrapping an entire
 * function definition, but we can't do much about that.
 */
#ifdef linux
// If we are compiling on Linux
#include <sys/sysinfo.h>

#elif defined __APPLE__ && defined __MACH__
// If we are compiling on Mac OS X
#include <sys/sysctl.h>

#else
// Default to just trying the tests regardless of memory
// We define the function as a macro below

#endif

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"


/**
 * @brief Check total system memory
 *
 * Used to make sure we're running on a system with
 * enough memory to actually execute the test.
 *
 * This function is not PERFECT, it's just intended to catch
 * some potential problems.
 *
 * @param required_memory amount of memory we need to have (in megs)
 *
 * @returns true if system memory is >= required memory
 **/
#ifdef linux
// We are compiling on Linux
static int system_memchk_megs_required (int required_megs)
{
	unsigned long required_memory = required_megs * 1024 * 1024;
	struct sysinfo sys;
	if (sysinfo (&sys) != 0)
	{
		fprintf (stderr, "Could not access system memory.  Assume not enough memory\n");
		return 0;
	}

	// sysinfo apparently doesnt work on 32bit systems with 4 gigs of RAM
	// It will return a very small amount of memory.
	// Let's just assume that a report of a very small amount of memory (say < 100 Mb)
	// means we actually have tons of memory.  Automatic success
	if (sys.totalram < 100*1024*1024)
	{
		return 1;
	}

	// Try to avoid overflow problems on systems with a lot of memory
	return sys.totalram > required_memory || (sys.totalswap > required_memory - sys.totalram);
}
#elif defined __APPLE__ && defined __MACH__
// We are compiling on Mac OS X
static int system_memchk_megs_required (int required_megs)
{
	unsigned long long required_memory = required_megs * 1024 * 1024;
	uint64_t totalmem;
	size_t sizeof_totalmem;

	sizeof_totalmem = sizeof(totalmem);

	// Read the hw.memsize sysctl value
	// hw.memsize will report the actual amount of physical memory because it's a 64 bit uint
	// hw.physmem is an int and will only report 2GB max
	if ( sysctlbyname("hw.memsize", &totalmem, &sizeof_totalmem, NULL, 0) != 0 ) {
		fprintf (stderr, "Could not access system memory.  Assume not enough memory\n");
		return 0;
	}

	// Shouldn't have to worry about overflow with unsigned long long amounts of memory
	// At least, not for a decade or two
	return totalmem > required_memory;
}
#else
// Default to just trying the tests regardless of memory
#define system_memchk_megs_required(arg) 1
#endif

static void recv_callback(struct client_socket* s)
{
	int len;
//	printf("%p: Data %s", s, s->rxbuffer);

	if ( strchr(s->rxbuffer, '\n')) {
		len = 0;
		while ( s->rxlen - len > 0 ) {
			socketserver_send(s, s->rxbuffer+len, 1);
			len++;
		}
		s->rxlen=0;
	}
}

static int redir_fd_number()
{
	FILE *fp;
	char number[1024];
	int fd_number = 0;

	/* Open the command for reading. */
	fp = popen("/usr/sbin/lsof -c redir | grep TCP | grep ESTABLISHED | wc -l", "r");
	if (fp == NULL) {
		printf("Failed to run command\n" );
		return fd_number;
	}

	if(fgets(number, sizeof(number) - 1, fp) != NULL) {
		fd_number = atoi(number);
	}

	/* close */
	pclose(fp);

	return fd_number;
}

static void single_redir_single_server_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	int j, increment, fd_number[10];
	char redir_command[100];
	int num, retry;

	// Require at least 1000 megs for this test
	CU_ASSERT_FATAL(system_memchk_megs_required (1000));

	// Create server
	//svr = socketserver_create(12345, connect_callback, recv_callback, disconnect_callback);
	svr = socketserver_create(12345, NULL, recv_callback, NULL);

	// 10 changes on connection number per server
	for(j = 0; j < 10; j++) {
		sprintf(redir_command, "../redir localhost:12345 -n%d", j * 10 + 100);
		fp = jopen(redir_command);
		if ( ! fp ) {
			fprintf(stderr, "Could not open ../redir\n");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			return;
		}

		// Perform 100 tests
		for(i=0; i<100; i++) {
			sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
			ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
			ret=read(fp->from_stdout, buffer, 1024);
			if ( ret >= 0 ) {
				buffer[ret]='\0';
			}

			if ( strcmp(buffer, sendbuffer) ) {
				fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
				CU_FAIL("We did not get the response we expected with enormous connections test - single redir sigle server");
				socketserver_signal_shutdown(svr);
				socketserver_waitfor_shutdown(svr);
				jclose(fp);
				return;
			}
		}
		fd_number[j] = redir_fd_number();
		for(retry = 0; retry < 100; retry++) {
			if(fd_number[j] == j * 10 + 100) {
				break;
			}
			usleep(1000000);
			fd_number[j] = redir_fd_number();
		}
		printf("Opened TCP file descriptor:%d, retry:%d\n", fd_number[j], retry);

		jclose(fp);

		num = redir_fd_number();
		for(retry = 0; retry < 200; retry++) {
			if(num == 0) {
				break;
			}
			usleep(1000000);
			num = redir_fd_number();
		}
		//printf("Wait for shutting down with retry:%d\n", retry);
	}

	increment = fd_number[1] - fd_number[0];
	for(j = 2; j < 10; j++) {
		if(fd_number[j] - fd_number[j - 1] > increment) {
			CU_FAIL("We did not have the correct file descriptor number with enormous connection test - single redir and sigle server");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			return;
		}
	}

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	CU_PASS("We have passed this test - enormous connections test with single redir and sigle server");
	return;
}

static void single_redir_multiple_server_test()
{
	int server_num = 3;
	struct socketserver* svr[server_num];
	JFILE* fp;
	int i, j, m;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	int increment, fd_number[10];
	char redir_command[1024];
	char hosts[1024];
	int num, retry;

	// Require at least 1000 megs for this test
	CU_ASSERT_FATAL(system_memchk_megs_required (1000));

	sprintf(hosts, "../redir");
	// Create multiple servers
	for(m = 0; m < server_num; m++) {
		sprintf(hosts, "%s localhost:%d", hosts, 12345 + m);
	}

	// 10 changes on connection number per server
	for(j = 0; j < 10; j++) {
		// Create multiple servers
		for(m = 0; m < server_num; m++) {
			svr[m] = socketserver_create(12345 + m, NULL, recv_callback, NULL);
		}

		sprintf(redir_command, "%s -n%d", hosts, j * 10 + 100);
		printf("Redir command:%s\n", redir_command);
		fp = jopen(redir_command);
		if ( ! fp ) {
			fprintf(stderr, "Could not open ../redir\n");
			for(m = 0; m < server_num; m++) {
				socketserver_signal_shutdown(svr[m]);
				socketserver_waitfor_shutdown(svr[m]);
			}
			return;
		}

		// Perform 100 tests
		for(i = 0; i < 100; i++) {
			sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
			ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
			ret=read(fp->from_stdout, buffer, 1024);
			if ( ret >= 0 ) {
				buffer[ret]='\0';
			}

			if ( strcmp(buffer, sendbuffer) ) {
				fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
				CU_FAIL("We did not get the response we expected with enormous connections test - single redir multiple servers");
				for(m = 0; m < server_num; m++) {
					socketserver_signal_shutdown(svr[m]);
					socketserver_waitfor_shutdown(svr[m]);
				}
				jclose(fp);
				return;
			}
		}
		fd_number[j] = redir_fd_number();
		for(retry = 0; retry < 100; retry++) {
			if(fd_number[j] == server_num * (j * 10 + 100)) {
				break;
			}
			usleep(1000000);
			fd_number[j] = redir_fd_number();
		}
		printf("Opened TCP file descriptor:%d, retry:%d\n", fd_number[j], retry);

		jclose(fp);
		for(m = 0; m < server_num; m++) {
			socketserver_signal_shutdown(svr[m]);
			socketserver_waitfor_shutdown(svr[m]);
		}

		num = redir_fd_number();
		for(retry = 0; retry < 200; retry++) {
			if(num == 0) {
				break;
			}
			usleep(1000000);
			num = redir_fd_number();
		}
		//printf("Wait for shutting down with retry:%d\n", retry);

	}

	increment = fd_number[1] - fd_number[0];
	for(j = 2; j < 10; j++) {
		if(fd_number[j] - fd_number[j - 1] > increment) {
			CU_FAIL("We did not have the correct file descriptor number with enormous connection test - single redir and multiple servers");
			return;
		}
	}

	CU_PASS("We have passed this test - enormous connections test with single redir and multiple servers");
	return;
}

static void multiple_redir_multiple_server_test()
{
	int server_num = 3;
	int redir_num = 3;
	struct socketserver* svr[server_num];
	JFILE* fp[redir_num];
	int i, j, m, n;
	char sendbuffer[256];
	char buffer[1024];
	int ret;
	int increment, fd_number[10];
	char redir_command[1024];
	char hosts[1024];  
	int num, retry;

	// Require at least 2000 megs for this test
	CU_ASSERT_FATAL(system_memchk_megs_required (2000));

	sprintf(hosts, "../redir");
	// Create multiple servers
	for(m = 0; m < server_num; m++) {
		sprintf(hosts, "%s localhost:%d", hosts, 12345 + m);
	}

	// 10 changes on connection number per server
	for(j = 0; j < 10; j++) {
		for(m = 0; m < server_num; m++) {
			svr[m] = socketserver_create(12345 + m, NULL, recv_callback, NULL);
		}

		sprintf(redir_command, "%s -n%d", hosts, j * 10 + 1);
		printf("Redir command:%s\n", redir_command);

		// Run multiple redirs
		for(n = 0; n < redir_num; n++) {
			fp[n] = jopen(redir_command);
			if ( ! fp[n] ) {
				fprintf(stderr, "Could not open ../redir\n");
				for(m = 0; m < server_num; m++) {
					socketserver_signal_shutdown(svr[m]);
					socketserver_waitfor_shutdown(svr[m]);
				}
				redir_num = n;
				for(n = 0; n < redir_num; n++) {
					jclose(fp[n]);
				}
				return;
			}
		}

		// Send requests to all redirs
		for(n = 0; n < redir_num; n++) {
			// Perform 100 tests
			for(i = 0; i < 100; i++) {
				sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
				ret = write(fp[n]->to_stdin, sendbuffer, strlen(sendbuffer));
				ret=read(fp[n]->from_stdout, buffer, 1023);
				if ( ret >= 0 ) {
					buffer[ret]='\0';
				}

				if ( strcmp(buffer, sendbuffer) ) {
					fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
					CU_FAIL("We did not get the response we expected with enormous connections test - multiple redir multiple servers");
					for(m = 0; m < server_num; m++) {
						socketserver_signal_shutdown(svr[m]);
						socketserver_waitfor_shutdown(svr[m]);
					}
					for(n = 0; n < redir_num; n++) {
						jclose(fp[n]);
					}
					return;
				}
			}

		}

		fd_number[j] = redir_fd_number();
		for(retry = 0; retry < 100; retry++) {
			if(fd_number[j] == server_num * redir_num * (j * 10 + 1)) {
				break;
			}
			usleep(1000000);
			fd_number[j] = redir_fd_number();
		}
		printf("Opened TCP file descriptor:%d, retry:%d\n", fd_number[j], retry);

		for(m = 0; m < server_num; m++) {
			socketserver_signal_shutdown(svr[m]);
			socketserver_waitfor_shutdown(svr[m]);
		}
		for(n = 0; n < redir_num; n++) {
			jclose(fp[n]);
		}

		num = redir_fd_number();
		for(retry = 0; retry < 200; retry++) {
			if(num == 0) {
				break;
			}
			usleep(1000000);
			num = redir_fd_number();
		}
		//printf("Wait for shutting down with retry:%d\n", retry);
	}

	increment = fd_number[1] - fd_number[0];
	for(j = 2; j < 10; j++) {
		if(fd_number[j] - fd_number[j - 1] > increment) {
			CU_FAIL("We did not have the correct file descriptor number with enormous connection test - multiple redir and multiple servers");
			
			return;
		}
	}

	CU_PASS("We have passed this test - enormous connections test with multiple redir and multiple servers");
	return;
}

static void multiple_redir_multiple_server_shutdown_test()
{
	int server_num = 10;
	int redir_num = 3;
	struct socketserver* svr[server_num];
	JFILE* fp[redir_num];
	int i, j, m, n;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	int decrement, fd_number[10];
	char redir_command[1024];
	char hosts[1024];
	int num, retry;

	// Require at least 2000 megs for this test
	CU_ASSERT_FATAL(system_memchk_megs_required (2000));

	sprintf(hosts, "../redir");
	// Create multiple servers
	for(m = 0; m < server_num; m++) {
		svr[m] = socketserver_create(12345 + m, NULL, recv_callback, NULL);
		sprintf(hosts, "%s localhost:%d", hosts, 12345 + m);
	}

	// Fix 30 connections for each server
	sprintf(redir_command, "%s -n30", hosts);
	printf("Redir command:%s\n", redir_command);

	// Run multiple redirs
	for(n = 0; n < redir_num; n++) {
		fp[n] = jopen(redir_command);
		if ( ! fp[n] ) {
			fprintf(stderr, "Could not open ../redir\n");
			for(m = 0; m < server_num; m++) {
				socketserver_signal_shutdown(svr[m]);
				socketserver_waitfor_shutdown(svr[m]);
			}
			redir_num = n;
			for(n = 0; n < redir_num; n++) {
				jclose(fp[n]);
			}
			return;
		}
	}

	// 10 changes on connection number per server
	// Shut down one server in each change, server_num must be greater than changes of 10
	for(j = 0; j < 10; j++) {
		// Send requests to all redirs
		for(n = 0; n < redir_num; n++) {
			// Perform 100 tests
			for(i = 0; i < 100; i++) {
				sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
				ret = write(fp[n]->to_stdin, sendbuffer, strlen(sendbuffer));
				ret=read(fp[n]->from_stdout, buffer, 1024);
				if ( ret >= 0 ) {
					buffer[ret]='\0';
				}

				if ( strcmp(buffer, sendbuffer) ) {
					fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
					CU_FAIL("We did not get the response we expected with enormous connections test - multiple redir multiple servers");
					for(m = 0; m < server_num; m++) {
						socketserver_signal_shutdown(svr[m]);
						socketserver_waitfor_shutdown(svr[m]);
					}
					for(n = 0; n < redir_num; n++) {
						jclose(fp[n]);
					}
					return;
				}
			}

		}

		fd_number[j] = redir_fd_number();
		for(retry = 0; retry < 100; retry++) {
			if(fd_number[j] == server_num * redir_num * 30) {
				break;
			}
			usleep(1000000);
			fd_number[j] = redir_fd_number();
		}
		printf("Opened TCP file descriptor:%d, retry:%d\n", fd_number[j], retry);

		// shut down one server
		socketserver_signal_shutdown(svr[server_num - 1]);
		socketserver_waitfor_shutdown(svr[server_num - 1]);
		printf("Shutting down server - localhost:%d\n", 12345 + server_num - 1);
		server_num--;

		num = redir_fd_number();
		for(retry = 0; retry < 200; retry++) {
			if(num == 0 || num == server_num * redir_num * 30) {
				break;
			}
			usleep(1000000);
			num = redir_fd_number();
		}
		//printf("Wait for shutting down with retry:%d\n", retry);
	}

	decrement = fd_number[0] - fd_number[1];
	for(j = 2; j < 10; j++) {
		if(fd_number[j - 1] - fd_number[j] > decrement) {
			for(m = 0; m < server_num; m++) {
				socketserver_signal_shutdown(svr[m]);
				socketserver_waitfor_shutdown(svr[m]);
			}
			for(n = 0; n < redir_num; n++) {
				jclose(fp[n]);
			}
			CU_FAIL("We did not have the correct file descriptor number with enormous connection test - multiple redir with shutting down server");
			
			return;
		}
	}

	for(m = 0; m < server_num; m++) {
		socketserver_signal_shutdown(svr[m]);
		socketserver_waitfor_shutdown(svr[m]);
	}
	for(n = 0; n < redir_num; n++) {
		jclose(fp[n]);
	}
	CU_PASS("We have passed this test - enormous connections test with multiple redir and multiple servers");
	return;
}

static void single_redir_single_server_max_connection_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	int j, fd_number[10];
	char redir_command[100];
	int num, retry, max_conn = 1014; 

	// Require at least 4000 megs for this test
	CU_ASSERT_FATAL(system_memchk_megs_required (4000));

	// Create server
	svr = socketserver_create(12345, NULL, recv_callback, NULL);

	sprintf(redir_command, "../redir localhost:12345 -n%d", max_conn);
	fp = jopen(redir_command);
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// 10 changes on connection number per server
	for(j = 0; j < 10; j++) {
		// Perform 1000 requests
		for(i=0; i<1000; i++) {
			sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
			ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
			ret=read(fp->from_stdout, buffer, 1024);
			if ( ret >= 0 ) {
				buffer[ret]='\0';
			}

			if ( strcmp(buffer, sendbuffer) ) {
				fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
				CU_FAIL("We did not get the response we expected with enormous connections test - single redir and sigle server with max connection");
				socketserver_signal_shutdown(svr);
				socketserver_waitfor_shutdown(svr);
				jclose(fp);
				return;
			}
		}
		fd_number[j] = redir_fd_number();
		printf("Opened TCP file descriptor:%d\n", fd_number[j]);
	}

	num = redir_fd_number();
	for(retry = 0; retry < 200; retry++) {
		if(num == max_conn) {
			break;
		}
		usleep(1000000);
		num = redir_fd_number();
	}
	//printf("Wait for shutting down with retry:%d\n", retry);
	
	if(num != max_conn) {
		CU_FAIL("We did not have the correct file descriptor number with enormous connection test - single redir and sigle server with max connection");
		jclose(fp);
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	jclose(fp);
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	CU_PASS("We have passed this test - enormous connections test with single redir and sigle server with max connection");
	return;
}

int enormous_connections_test_init()
{
	CU_pSuite pSuite = NULL;

	pSuite = CU_add_suite("Enormous Conections Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// single redir with single server
	if ( (NULL == CU_add_test(pSuite, "Enormous Conections Test - Single Redir with Single Server\n", single_redir_single_server_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// single redir with multiple servers
	if ( (NULL == CU_add_test(pSuite, "Enormous Conections Test - Single Redir with Multiple Server\n", single_redir_multiple_server_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// multiple redir with multiple servers
	if ( (NULL == CU_add_test(pSuite, "Enormous Conections Test - Multiple Redir with Multiple Server\n", multiple_redir_multiple_server_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// multiple redir with shutting down servers
	if ( (NULL == CU_add_test(pSuite, "Enormous Conections Test - Multiple Redir with Shutting down Server\n", multiple_redir_multiple_server_shutdown_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// multiple redir with shutting down servers
	if ( (NULL == CU_add_test(pSuite, "Enormous Conections Test - Single Redir and Single Server with max connections\n", single_redir_single_server_max_connection_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
