/**
 * @brief The following makes sure we can properly filtering http/https based requests
 *
 * Most Web Caches can handle NON-HTTP protocols.
 * This is based on http://www.ietf.org/rfc/rfc2817.txt
 *
 * HTTPS/Connection Proxy
 *  www.site.com ip/fqdn username CONNECT
 * Website Based Request
 *  http://www.site.com ip/fqdn username GET
 *  http://www.site.com ip/fqdn username POST
 *  http://www.site.com ip/fqdn username ...
 *
 * Note: squid may send different number of fields based on redir format.
 * This makes sure the Squid 2.5 format works properly.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>

#include "CUnit/Basic.h"
#include "jopen.h"
#include "socketserver.h"



static void connect_callback(__attribute__ ((unused)) struct client_socket* s)
{
}
static void recv_deny_callback(struct client_socket* s)
{
	if ( strchr(s->rxbuffer, '\n') ) {
		printf("ECHO %s\n", s->rxbuffer);
		socketserver_send(s, "denied.site\n", 12);
		s->rxlen=0;
	}
}
static void recv_allow_callback(struct client_socket* s)
{
	if ( strchr(s->rxbuffer, '\n') ) {
		printf("ECHO %s\n", s->rxbuffer);
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}
}
static void recv_noop_callback(struct client_socket* s)
{
	s->rxlen=0;
}
static void disconnect_callback(__attribute__ ((unused)) struct client_socket* s)
{
}


static int send_and_get_other(JFILE* fp, char* sbuffer, char* ibuffer)
{
	char sendbuffer[1024];
	char buffer[1024];
	int ret;
	char* ptr;

	strcpy(sendbuffer, sbuffer);
	strcat(sendbuffer, " -/- - CONNECT\n");

	ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

	int len=0;
	buffer[0]='\0';
	while ( strchr(buffer, '\n') == NULL ) {
		ret=read(fp->from_stdout, buffer+len, 1024);
		if ( ret > 0 ) {
			len+=ret;
			buffer[len]='\0';
		}
	}

	// copy what we are looking for into the sendbuffer
	// reused the sendbuffer
	strcpy(sendbuffer, ibuffer);
	ptr=strchr(sendbuffer, ' ');
	if ( ptr ) {
		*ptr='\0';
	}
	ptr=strchr(buffer, ' ');
	if ( ptr ) {
		*ptr='\0';
	}
	ptr=strchr(buffer, '\n');
	if ( ptr ) {
		*ptr='\0';
	}

	if ( strcmp(buffer, sendbuffer) ) {
		fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
		CU_FAIL("We did not get the response we expected");
		return 0;
	}


	return 1;
}

static int send_and_get(JFILE* fp, char* sbuffer)
{
	return send_and_get_other(fp, sbuffer, sbuffer);
}


/**
 * @brief Test https allow
 */
void https_allow_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11225, connect_callback, recv_allow_callback, disconnect_callback);

	fp = jopen("../redir localhost:11225 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// send and get something
	// this should work
	CU_ASSERT(send_and_get(fp, "google.ca"));
	CU_ASSERT(send_and_get(fp, "google.ca"));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief Test https deny
 */
void https_deny_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11225, connect_callback, recv_deny_callback, disconnect_callback);

	fp = jopen("../redir localhost:11225 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// send and get something
	// this should work
	CU_ASSERT(send_and_get_other(fp, "google.ca", "denied.site"));
	CU_ASSERT(send_and_get_other(fp, "google.ca", "denied.site"));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief Test https fail open
 */
void https_failopen_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11225, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir localhost:11225");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// send and get something
	// this should work
	CU_ASSERT(send_and_get(fp, "google.ca:123"));
	CU_ASSERT(send_and_get(fp, "google.ca:123"));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief Test https fail open
 */
void https_failclose_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11225, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir localhost:11225 -f http://failclose");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// The noop call back does nothing, so these will also fail
	CU_ASSERT(send_and_get_other(fp, "google.ca", "failclose:443"));
	CU_ASSERT(send_and_get_other(fp, "google.ca:443", "failclose:443"));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	CU_ASSERT(send_and_get_other(fp, "google.ca", "failclose:443"));
	CU_ASSERT(send_and_get_other(fp, "google.ca", "failclose:443"));

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief Test https fail open different host
 */
void https_failclose_host_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11225, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir localhost:11225 -f http://failclose -s https.com");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// noop callback does nothing, so these will fail
	CU_ASSERT(send_and_get_other(fp, "google.ca", "https.com:443"));
	CU_ASSERT(send_and_get_other(fp, "google.ca:123", "https.com:443"));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	CU_ASSERT(send_and_get_other(fp, "google.ca", "https.com:443"));
	CU_ASSERT(send_and_get_other(fp, "google.ca", "https.com:443"));

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}


/**
 * @brief Test https fail open different host and port
 */
void https_failclose_port_test()
{
	struct socketserver* svr;
	JFILE* fp;

	// Create a new amazing server
	svr = socketserver_create(11225, connect_callback, recv_noop_callback, disconnect_callback);

	fp = jopen("../redir localhost:11225 -f http://failclose -s https.com:123");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	// noop callback does nothing, so this will fail close
	CU_ASSERT(send_and_get_other(fp, "google.ca", "https.com:123"));
	CU_ASSERT(send_and_get_other(fp, "google.ca:123", "https.com:123"));

	// shutdown the server
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	CU_ASSERT(send_and_get_other(fp, "google.ca", "https.com:123"));
	CU_ASSERT(send_and_get_other(fp, "google.ca", "https.com:123"));

	// so will this
	jclose(fp);

	CU_PASS("We have passed this test");

	return;
}

/**
 * @brief This setup and creation of this most amazing test
 *
 */
int https_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test1
	pSuite = CU_add_suite("HTTPS Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "HTTPS Allowed", https_allow_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "HTTPS Denied", https_deny_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "HTTPS Fail Open", https_failopen_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "HTTPS Fail Close", https_failclose_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "HTTPS Fail Close Different Host", https_failclose_host_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	if ( (NULL == CU_add_test(pSuite, "HTTPS Fail Close Different Port", https_failclose_port_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
