/**
 * @file
 * @brief This file tests having multiple servers with various errors
 *
 *  Do we load balancing properly between two servers.
 *  Do we work properly when one server does not exist.
 *  Do we work when connecting to one server twice.
 *
 *  When one server keeps disconnecting after a request, does the other server properly handle the requests.
 *  When one server never sends a reply, do we utilize the other server properly.
 *  When one server never sends a \n, do we utilize the other server properly.
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

struct conninfo {
	int total_count;
	int connections;
	int* counters;
};
static struct conninfo* conninfo_init()
{
	struct conninfo* ci;
	ci = malloc(sizeof(struct conninfo));
	ci->total_count=0;
	ci->connections=0;
	ci->counters=NULL;
	return ci;
}
static void conninfo_free(struct conninfo* ci)
{
	if ( ci->counters ) {
		free(ci->counters);
	}
	free(ci);
}

static void connect_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;

	// our current connection id in the counters array
	s->tag = (void*)ci->connections;
	// increment, then reallocate
	ci->connections++;
	ci->counters =  realloc(ci->counters, sizeof(int) * ci->connections);
	ci->counters[(int)s->tag]=0;
	printf("%p: Connected number %d\n", s, (int)s->tag);
}

/**
 * @brief This call back will increment our conninfo information and allow the request
 */
static void recv_allow_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;

		ci->total_count++;
		ci->counters[(int)s->tag]++;
	}
}
/**
 * @brief This will do the same as recv_allow_callback but will also disconnect
 */
static void recv_disconnect_callback(struct client_socket* s)
{
	// do the allow callback
	recv_allow_callback(s);

	// disconnect
	socketserver_connection_shutdown(s);
}

/**
 * @brief This will not send any response and will just disconnect
 */
static void recv_onlydisconnect_callback(struct client_socket* s)
{
	// do not allow or deny, but consume the data
	s->rxlen=0;

	// disconnect
	socketserver_connection_shutdown(s);
}

/**
 * @brief This will simply not send anything, noop
 */
static void recv_noop_callback(struct client_socket* s)
{
	// do nothing... but consume the data
	s->rxlen=0;
}

/**
 * @brief This sends some garbage data back with no \n
 */
static void recv_garbage_callback(struct client_socket* s)
{
	// send some garbage
	socketserver_send(s, "welcome, this is the farm house", 10);
	// consume
	s->rxlen=0;
}

static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}

/**
 * @brief Create two servers and send 100 requests to them
 *
 * This test creates two servers
 * We generate 100 requests
 * We make sure the requests are distributed between the two servers
 * Fails if
 *  - Any request fails
 *  - A server gets more then 80 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_simple_twoserver_test()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_allow_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir -f http://failcloseurl -e --log-level=3 localhost:12345 localhost:12346");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return;
	}
	printf("jopen return OK\n");

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);

	// We processed to many requests on a single server
	if ( ci1->total_count + ci2->total_count != 100 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("One server instance processed more then 80 requests.  Distribution not very good");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The Multiple Server Redir test has passed");
	return;
}


/**
 * @brief Create only 1 server, load balancing between two servers
 *
 * This test creates one server
 * We generate 100 requests
 * All requests should goto the one server that exists
 *
 * Fails if
 *  - Any request fails
 *  - The one server does not get 100 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_1up_1down_test()
{
	struct socketserver* svr1;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);

	ci1 = (struct conninfo*)svr1->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);

	// We processed to many requests on a single server
	if ( ci1->total_count != 100 ) {
		socketserver_waitfor_shutdown(svr1);
		jclose(fp);
		CU_FAIL("The one alive server did not get 100 requests");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	jclose(fp);

	CU_PASS("The one server got all requests and processed them correctly");
	return;
}

/**
 * @brief Create only 1 server, connect to this server twice
 *
 * This test creates one server
 * We connect to this server 2 times
 * We generate 100 requests
 * All requests should goto the one server that exists
 * We check the connection distribution
 *
 * Fails if
 *  - Any request fails
 *  - The one server does not get 100 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_1up_2connections_test()
{
	struct socketserver* svr1;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir -n 2 localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);

	ci1=svr1->tag;
	printf("Server 1 Processed: %d requests on %d connections\n", ci1->total_count, ci1->connections);

	// we must have processed 100 requests
	// on 2 connections
	// and those connections must have processed < 80 requests each (usually this is a 50/50)
	if ( ci1->total_count != 100 || ci1->connections != 2 || ci1->counters[0] > 80 || ci1->counters[1] > 80 ) {
		socketserver_waitfor_shutdown(svr1);
		jclose(fp);
		CU_FAIL("The one alive server did not get 100 requests");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	jclose(fp);

	CU_PASS("The one server got all requests and processed them correctly");
	return;
}


/**
 * @brief Create two servers and send 100 requests to them, one server always disconnects
 *
 * This test creates two servers
 *   - One of these servers disconnects after a request
 * We generate 100 requests
 * We make sure the requests are distributed between the two servers
 * Fails if
 *  - Any request fails
 *  - A server gets more then 80 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_2servers_1disconnects_test()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_disconnect_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);

	// We processed to many requests on a single server
	if ( ci1->total_count + ci2->total_count != 100 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("The amount of processed requests is off for some reason.");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The Multiple Server Redir test has passed");
	return;
}


/**
 * @brief Create two servers and send 100 requests to them, one server never replies
 *
 * This test creates two servers
 *   - One of these servers simply consumes all data
 * We generate 100 requests
 * We make sure the requests are distributed between the two servers
 * Fails if
 *  - Any request fails
 *  - A server gets more then 80 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_2servers_1noop_test()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_noop_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);

	// We processed to many requests on a single server
	if ( ci1->total_count != 100 || ci2->total_count != 0 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("Only server 1 should have processed anything.");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The multiple server test where one server is dead has passed");
	return;
}

/**
 * @brief Create two servers and send 100 requests to them, one server sends garbage data
 *
 * This test creates two servers
 *   - One of these servers simply consumes all data and sends some garbage back
 * We generate 100 requests
 * We make sure the requests are distributed between the two servers
 * Fails if
 *  - Any request fails
 *  - A server gets more then 80 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_2servers_1garbage_test()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_garbage_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);

	// We processed to many requests on a single server
	if ( ci1->total_count != 100 || ci2->total_count != 0 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("Only server 1 should have processed anything.");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The multiple server test where one server is dead has passed");
	return;
}

/**
 * @brief Create three servers, send 100 requests to them, and have two fail
 * after their first request
 *
 * This test creates three servers
 * We generate 100 requests
 * We make sure that when two of the servers go down in sequence we don't lose
 * requests
 * Fails if
 *  - Any request fails
 *  - We don't process all 100 requests
 *
 * Uses the recv_allow_callback and recv_disconnect_callback
 *
 */
static void multiserver_3servers_2disconnect()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	struct socketserver* svr3;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;
	struct conninfo* ci3;

	// Create three servers: one always allows, the other two disconnect after every request
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_disconnect_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);
	svr3 = socketserver_create(12347, connect_callback, recv_disconnect_callback, disconnect_callback);
	svr3->tag = conninfo_init();
	socketserver_tagfree_func(svr3, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 localhost:12347 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_signal_shutdown(svr3);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		socketserver_waitfor_shutdown(svr3);
		return;
	}

	// Perform 100 tests
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// Truncate sent at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// Truncate received buffer at first newline
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// Truncate received buffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_signal_shutdown(svr3);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			socketserver_waitfor_shutdown(svr3);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);
	socketserver_signal_shutdown(svr3);

	ci1=svr1->tag;
	ci2=svr2->tag;
	ci3=svr3->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);
	printf("Server 3 Processed: %d requests\n", ci3->total_count);

	// We didn't process the right number of requests
	if ( ci1->total_count + ci2->total_count + ci3->total_count != 100 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		socketserver_waitfor_shutdown(svr3);
		jclose(fp);
		CU_FAIL("The amount of processed requests is off for some reason");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	socketserver_waitfor_shutdown(svr3);
	jclose(fp);

	CU_PASS("The Multiple Server cascading fail Redir test has passed");
	return;
}

/**
 * @brief Create three servers, send 100 requests to them, and have two
 * disconnect immediately, without sending any response
 * 
 *
 * This test creates three servers
 * We generate 100 requests
 * We make sure that when two of the servers constantly disconnect without
 * repsonse, we don't lose requests
 * Fails if
 *  - Any request fails
 *  - We don't process all 100 requests
 *
 * Uses the recv_allow_callback and recv_onlydisconnect_callback
 *
 */
static void multiserver_3servers_2onlydisconnect()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	struct socketserver* svr3;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;
	struct conninfo* ci3;

	// Create three servers: one always allows, the other two send no response and just disconnect
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_onlydisconnect_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);
	svr3 = socketserver_create(12347, connect_callback, recv_onlydisconnect_callback, disconnect_callback);
	svr3->tag = conninfo_init();
	socketserver_tagfree_func(svr3, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s() Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 localhost:12347 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_signal_shutdown(svr3);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		socketserver_waitfor_shutdown(svr3);
		return;
	}

	// Perform 100 tests
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// Truncate sent at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// Truncate received buffer at first newline
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// Truncate received buffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_signal_shutdown(svr3);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			socketserver_waitfor_shutdown(svr3);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);
	socketserver_signal_shutdown(svr3);

	ci1=svr1->tag;
	ci2=svr2->tag;
	ci3=svr3->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);
	printf("Server 3 Processed: %d requests\n", ci3->total_count);

	// We didn't process the right number of requests
	if ( ci1->total_count + ci2->total_count + ci3->total_count != 100 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		socketserver_waitfor_shutdown(svr3);
		jclose(fp);
		CU_FAIL("The amount of processed requests is off for some reason");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	socketserver_waitfor_shutdown(svr3);
	jclose(fp);

	CU_PASS("The Multiple Server two servers just disconnecting Redir test has passed");
	return;
}



/**
 * @brief Create two servers and send 100 requests to them
 *
 * This test creates two servers
 * We generate 100 requests
 * We make sure the requests are distributed between the two servers
 * Fails if
 *  - Any request fails
 *  - A server gets more then 80 requests
 *
 * Uses the recv_allow_callback
 *
 */
static void multiserver_many_connections_fail()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_allow_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir -n 10 localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		if (i % 2) {
			socketserver_signal_reset_connections(svr1);
			socketserver_waitfor_reset_connections(svr1);
		}

		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The Close Many connections test is passed");
	return;
}




int multiple_server_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test1
	pSuite = CU_add_suite("Multiple Server Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Two servers, load balance properly
	if ( (NULL == CU_add_test(pSuite, "Load Balance Two Server Test", multiserver_simple_twoserver_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "One Server Up One Server Always Down", multiserver_1up_1down_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "One Server Up Two Connections", multiserver_1up_2connections_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Two Servers, one disconnects", multiserver_2servers_1disconnects_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Two Servers, one does nothing", multiserver_2servers_1noop_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Two Servers, one garbage response", multiserver_2servers_1garbage_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Three servers, two disconnect", multiserver_3servers_2disconnect)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Three servers, two only disconnect", multiserver_3servers_2onlydisconnect)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Two servers, 10 connections, one fails", multiserver_many_connections_fail)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
