/**
 * @file
 * @brief Tests to make sure the redirector can read request from squid and then send the 
 * request to nsp properly
 *
 * Request can be send in the folowing mode:
 * One test sends it in a an entire chunk
 * Another test sends it byte by byte
 * Another test sends it chunk by chunk randoly
 * Another test adds a random delay between each byte
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include "jopen.h"

#include "CUnit/Basic.h"

#include "socketserver.h"

static void connect_callback(struct client_socket* s)
{
	printf("%p: Connected\n", s);
}
static void recv_callback(struct client_socket* s)
{
	int len;
	printf("%p: Data %s", s, s->rxbuffer);


	if ( strchr(s->rxbuffer, '\n') ) {
		len = 0;
		while ( s->rxlen - len > 0 ) {
			socketserver_send(s, s->rxbuffer+len, 1);
			len++;
		}
		s->rxlen=0;
	}
}
static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}

void squid_entire_chunk_input_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	char expbuffer[256];
	int ret;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	// squid 2.5 input
	fp = jopen("../redir localhost -i squid2.5 -o squid2.5 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - squid sending entire chunk input");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	jclose(fp);
	CU_PASS("We have passed this test - squid sending entire chunk input");

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	return;
}

void squid_byte_by_byte_input_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	char expbuffer[256];
	int ret;
	int totallen, len;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	// squid 2.5 input
	fp = jopen("../redir localhost -i squid2.5 -o squid2.5 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i);

		totallen = strlen(sendbuffer);
		for(len = 0; len < totallen; len++) {
			write(fp->to_stdin, sendbuffer + len, 1);
		}

		len=0, ret=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - squid sending byte by byte input");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	jclose(fp);
	CU_PASS("We have passed this test - squid sending byte by byte input");

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	return;
}

void squid_random_delays_input_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	char expbuffer[256];
	int ret;
	int totallen, delay, len;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	// squid 2.5 input
	fp = jopen("../redir localhost -i squid2.5 -o squid2.5 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i);

		totallen = strlen(sendbuffer);
		for(len = 0; len < totallen; len++) {
			delay = random() % 10000;
			usleep(delay);
			write(fp->to_stdin, sendbuffer + len, 1);
		}

		len=0, ret=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - squid sending random delays input");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	jclose(fp);
	CU_PASS("We have passed this test - squid sending random delays input");

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	return;
}

void squid_random_chunks_input_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	char expbuffer[256];
	int ret;
	int totallen, tosend, len, delay;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	// squid 2.5 input
	fp = jopen("../redir localhost -i squid2.5 -o squid2.5 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i);

		totallen = strlen(sendbuffer);
		len = 0, ret = 0;
		while(len < totallen) {
			tosend = (random() % 5) + 1;
			delay = random() % 100000;
			usleep(delay);
			if(tosend > totallen - len) {
				tosend = totallen - len;
			}
			ret = write(fp->to_stdin, sendbuffer + len, tosend);
			len += ret;
		}

		len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - squid sending random chunks input");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	jclose(fp);
	CU_PASS("We have passed this test - squid sending random chunks input");

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	return;
}

int redir_squid_send_test_init()
{
	CU_pSuite pSuite = NULL;

	pSuite = CU_add_suite("Squid Request Sending Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Squid Sends Entire Chunk Input\n", squid_entire_chunk_input_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Squid Sends Byte by Byte Input\n", squid_byte_by_byte_input_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Squid Sends Random Chunks Input\n", squid_random_chunks_input_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Squid Sends Random Delays Input\n", squid_random_delays_input_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
